#include "ntddk.h"
#include "a1216e.h"


/*********************************************************************************
*
*		DRIVER ENTRY
*
**********************************************************************************/

NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath
    )
{
    PDEVICE_OBJECT deviceObject = NULL;
    NTSTATUS status;
    UNICODE_STRING uniNtNameString;
    UNICODE_STRING uniWin32NameString;
    KAFFINITY Affinity;
    ULONG MappedVector;
    PA1216E_DEVICE_EXTENSION extension;

	// Create dispatch points for create/open, close, unload, and ioctl
	DriverObject->MajorFunction[IRP_MJ_CREATE] = A1216EOpenClose;
	DriverObject->MajorFunction[IRP_MJ_CLOSE] = A1216EOpenClose;
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = A1216EDeviceControl;
	DriverObject->DriverUnload  = A1216EUnload;

    // Create counted string version of our device name.
    RtlInitUnicodeString( &uniNtNameString, NT_DEVICE_NAME );

    // Create the device object
    status = IoCreateDevice(DriverObject,
							sizeof(A1216E_DEVICE_EXTENSION),
							&uniNtNameString,
							FILE_DEVICE_KRNLDRVR,
							0,
							FALSE,
							&deviceObject
							);

    if ( NT_SUCCESS(status) )
    {
		// fill in the device extension
		extension = (PA1216E_DEVICE_EXTENSION) deviceObject->DeviceExtension;
		extension->DeviceObject = deviceObject;
		extension->InterruptConnected = FALSE;
		extension->BaseAddress = 0;
		extension->NumPoints = extension->PointsRemaining = 0;

		// tell I/O Manager we want it to read/write data via buffers (NOT relevant to DeviceIOControl)
		deviceObject->Flags |= DO_BUFFERED_IO;

		// register DPC routine
		// not this time
//		IoInitializeDpcRequest(deviceObject, A1216EDpcForIsrRoutine);
		
        // Create counted string version of our Win32 device name.
        RtlInitUnicodeString( &uniWin32NameString, DOS_DEVICE_NAME );
    
        // Create a link from our device name to a name in the Win32 namespace.
        status = IoCreateSymbolicLink( &uniWin32NameString, &uniNtNameString );

        if (!NT_SUCCESS(status))
        {
			IoDeleteDevice( DriverObject->DeviceObject );
        }
    }
    return status;
}

/*********************************************************************************
*
*		OPEN / CLOSE (nothing happens so they're the same)
*
**********************************************************************************/

NTSTATUS
A1216EOpenClose(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
{
    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest( Irp, IO_NO_INCREMENT );

    return STATUS_SUCCESS;
}

/*********************************************************************************
*
*		UNLOAD
*
**********************************************************************************/

VOID
A1216EUnload(
    IN PDRIVER_OBJECT DriverObject
    )
{
    UNICODE_STRING uniWin32NameString;
    PA1216E_DEVICE_EXTENSION extension;

    extension = DriverObject->DeviceObject->DeviceExtension;

    // Disconnect the interrupt
    if (extension->InterruptConnected)
		IoDisconnectInterrupt(extension->InterruptObject);

	// Free up the memory
	if (extension->DataBufferAllocated)
		ExFreePool(extension->DataBuffer);
	if (extension->ChannelTableAllocated)
		ExFreePool(extension->ChannelTable);

    // Create counted string version of our Win32 device name.
    RtlInitUnicodeString( &uniWin32NameString, DOS_DEVICE_NAME );

    // Delete the link from our device name to a name in the Win32 namespace.
    IoDeleteSymbolicLink( &uniWin32NameString );

    // Finally delete our device object
    IoDeleteDevice( DriverObject->DeviceObject );
}

/*********************************************************************************
*
*		DEVICE CONTROL
*
**********************************************************************************/

NTSTATUS
A1216EDeviceControl(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
{
    NTSTATUS status, ioConnectStatus;
    PIO_STACK_LOCATION irpSp;
    PA1216E_DEVICE_EXTENSION extension;
    ULONG ioControlCode;
    KIRQL irql;
    KAFFINITY Affinity;
    ULONG MappedVector;
	PULONG pIOBuffer;
	ULONG InBufferSize, OutBufferSize;
	ULONG nPort, n;
	USHORT BusNumber;
	USHORT timeout=0xFFFF;
	SHORT BusType;
	PSHORT NewTable;

    extension = DeviceObject->DeviceExtension;

    irpSp = IoGetCurrentIrpStackLocation( Irp );
	Irp->IoStatus.Information = 0;

    // Save the control code in the extension so we can use it later
    ioControlCode = irpSp->Parameters.DeviceIoControl.IoControlCode;

    switch (ioControlCode)
	{
		case IOCTL_A1216E_STOP:
			extension->PointsRemaining = 0;
			status = STATUS_SUCCESS;
			break;

		case IOCTL_A1216E_STATUS:
		    OutBufferSize = irpSp->Parameters.DeviceIoControl.OutputBufferLength;
			if (OutBufferSize < sizeof(LONG))
			{
				status = STATUS_INVALID_PARAMETER;
				break;
			}
			
			pIOBuffer = (PULONG)Irp->AssociatedIrp.SystemBuffer;
			*(PLONG)pIOBuffer = extension->PointsRemaining;
			Irp->IoStatus.Information = sizeof(LONG);
			status = STATUS_SUCCESS;
			break;

		case IOCTL_A1216E_GETDATA:
//			if (!extension->InterruptConnected)
//			{
//				status = STATUS_INVALID_DEVICE_REQUEST;
//				break;
//			}
		    OutBufferSize = irpSp->Parameters.DeviceIoControl.OutputBufferLength;
			if (OutBufferSize < extension->NumPoints * 8)
			{
				status = STATUS_INVALID_PARAMETER;
				break;
			}

			pIOBuffer = (PULONG)Irp->AssociatedIrp.SystemBuffer;
			for (n=0; n<extension->NumPoints; n++)
				((PULONG)pIOBuffer)[n*2] = extension->DataBuffer[n];
			Irp->IoStatus.Information = extension->NumPoints * 8;
			status = STATUS_SUCCESS;
			break;

		case IOCTL_A1216E_INIT_GAIN_TABLE:
			InBufferSize = irpSp->Parameters.DeviceIoControl.InputBufferLength;
			pIOBuffer = (PULONG)Irp->AssociatedIrp.SystemBuffer;

			for (n=0; n<sizeof(extension->GainTable); n++)
				extension->GainTable[n] = 0;

			RtlCopyBytes(extension->GainTable, pIOBuffer, InBufferSize<sizeof(extension->GainTable)?InBufferSize:sizeof(extension->GainTable));
			status = STATUS_SUCCESS;
	
			Irp->IoStatus.Information = 0;
			break;
		
		case IOCTL_A1216E_INIT_CHANNEL_TABLE:
			InBufferSize = irpSp->Parameters.DeviceIoControl.InputBufferLength;
			pIOBuffer = (PULONG)Irp->AssociatedIrp.SystemBuffer;

			if (extension->ChannelTable != NULL)
			{
				ExFreePool(extension->ChannelTable);
				extension->ChannelTableAllocated = FALSE;
			}
			extension->ChannelTable = (PSHORT)ExAllocatePool(NonPagedPool, InBufferSize);
			if (extension->ChannelTable == NULL)
			{
				status = STATUS_INSUFFICIENT_RESOURCES;
				break;
			}
			else
				extension->ChannelTableAllocated = TRUE;

			n = InBufferSize / 2;
			extension->NumChannelEntries = 0;
			while (extension->NumChannelEntries < n)
			{
				extension->ChannelTable[extension->NumChannelEntries] = ((PSHORT)pIOBuffer)[extension->NumChannelEntries];
				if (extension->ChannelTable[extension->NumChannelEntries] < 0)
					break;
				else
					extension->NumChannelEntries++;
			}
			if (extension->NumChannelEntries == 0)
			{
				ExFreePool(extension->ChannelTable);
				extension->ChannelTableAllocated = FALSE;
				status = STATUS_INVALID_PARAMETER;
				break;
			}
			// if there are entries after -1, resize our memory to optimize
			if (n - 1 > extension->NumChannelEntries)
			{
				NewTable = (PSHORT)ExAllocatePool(NonPagedPool, (extension->NumChannelEntries + 1) * 2);
				if (NewTable != NULL)
				{
					ExFreePool(extension->ChannelTable);
					extension->ChannelTable = NewTable;
				}
			}
			status = STATUS_SUCCESS;
			Irp->IoStatus.Information = 0;
			break;
		
		case IOCTL_A1216E_INIT_DRIVER:
		    InBufferSize = irpSp->Parameters.DeviceIoControl.InputBufferLength;
			if (InBufferSize < sizeof(A1216E_INIT_DATA))
			{
				status = STATUS_INVALID_PARAMETER;
				ExFreePool(extension->ChannelTable);
				extension->ChannelTableAllocated = FALSE;
				break;
			}

			if (!extension->ChannelTableAllocated)
			{
				status = STATUS_INVALID_DEVICE_REQUEST;
				break;
			}
			if (extension->InterruptConnected)
			{
				IoDisconnectInterrupt(extension->InterruptObject);
			}
			extension->InterruptConnected = FALSE;

			pIOBuffer = (PULONG)Irp->AssociatedIrp.SystemBuffer;
			irql = extension->IRQ = ((PA1216E_INIT_DATA)pIOBuffer)->IRQ;
			BusType = ((PA1216E_INIT_DATA)pIOBuffer)->BusType;
			BusNumber = ((PA1216E_INIT_DATA)pIOBuffer)->BusNumber;
			extension->BaseAddress = ((PA1216E_INIT_DATA)pIOBuffer)->BaseAddress;
			extension->NumPoints = ((PA1216E_INIT_DATA)pIOBuffer)->NumPoints;
			extension->PointsRemaining = extension->NumPoints;
			if (extension->NumPoints < 1)
			{
				ExFreePool(extension->ChannelTable);
				extension->ChannelTableAllocated = FALSE;
				status = STATUS_INVALID_PARAMETER;
				break;
			}
			// allocate buffer to store Data
			if (extension->DataBuffer != NULL)
			{
				ExFreePool(extension->DataBuffer);
				extension->DataBufferAllocated = FALSE;
			}
			extension->DataBuffer = (PULONG)ExAllocatePool(NonPagedPool, extension->NumPoints * 4);
			if (extension->DataBuffer == NULL)
			{
				ExFreePool(extension->ChannelTable);
				extension->ChannelTableAllocated = FALSE;
				status = STATUS_INSUFFICIENT_RESOURCES;
				break;
			}
			else
				extension->DataBufferAllocated = TRUE;

			RtlZeroMemory(extension->DataBuffer, extension->NumPoints * 4);

			nPort = extension->BaseAddress + 0xF;
			WRITE_PORT_UCHAR((PUCHAR)nPort, 0xB4); // stop counters
			
			nPort = extension->BaseAddress;
			WRITE_PORT_UCHAR((PUCHAR)nPort, 0xEA); // set up card
			extension->IntDisabled = FALSE;

			nPort = extension->BaseAddress + 1;
			READ_PORT_UCHAR((PUCHAR)nPort); // clear latch
			extension->IRQCleared = FALSE;

			//
			// This call will map our IRQ to a system vector. It will also fill
			// in the IRQL (the kernel-defined level at which our ISR will run),
			// and affinity mask (which processors our ISR can run on).
			//
			// We need to do this so that when we connect to the interrupt, we
			// can supply the kernel with this information.
			//
			MappedVector = HalGetInterruptVector(
				BusType,	// Interface type
				BusNumber,	// Bus number
				irql, // Bus interrupt level
				irql, // Bus interrupt vector
				&irql,	// IRQ level
				&Affinity	// Affinity mask
				);
			ioConnectStatus = IoConnectInterrupt(
				&extension->InterruptObject,
				A1216EInterruptServiceRoutine,
				extension->DeviceObject,
				NULL,
				MappedVector,
				irql,
				irql,
				Latched,
				FALSE,
				Affinity,
				FALSE
				);
			status = ioConnectStatus;

			if ( NT_SUCCESS (status) )
			{
				extension->InterruptConnected = TRUE;
			}
			
			// set the first channel in the table
			extension->Index = 0;
			A1216E_SetChannel(extension->ChannelTable[extension->Index], extension);
			break;
			
		default:
			status = STATUS_INVALID_DEVICE_REQUEST;
			break;
	}

	Irp->IoStatus.Status = status;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);
	return status;
}

/*********************************************************************************
*
*		SET CHANNEL
*
**********************************************************************************/

VOID A1216E_SetChannel(IN LONG Ch, IN PA1216E_DEVICE_EXTENSION extension)
{
	UCHAR Value;
	ULONG nPort;

	Value  = (Ch & 0xF);  // set MA0-MA3 (a/d channel)
	Value |= ((extension->GainTable[Ch] & 0x3) << 4);  // set GN0-GN1

	nPort = extension->BaseAddress + 2;
	WRITE_PORT_UCHAR((PUCHAR)nPort, Value);
}

/*********************************************************************************
*
*		INTERRUPT SERVICE ROUTINE
*
**********************************************************************************/

BOOLEAN
A1216EInterruptServiceRoutine(
    IN PKINTERRUPT Interrupt,
    IN OUT PVOID Context
    )
{
    PDEVICE_OBJECT DeviceObject;
    PA1216E_DEVICE_EXTENSION extension;
	ULONG nPort;
	ULONG Value;

    DeviceObject = Context;
    extension = DeviceObject->DeviceExtension;

	if (!extension->IRQCleared)
	{
		nPort = extension->BaseAddress + 1;
		READ_PORT_UCHAR((PUCHAR)nPort);
		extension->IRQCleared = TRUE;
		return TRUE;
	}

	if (extension->IRQ >= 8 && extension->IRQ <= 15)
	{
		nPort = 0xA0;
		WRITE_PORT_UCHAR((PUCHAR)nPort, 0x20);
	}
	nPort = 0x20;
	WRITE_PORT_UCHAR((PUCHAR)nPort, 0x20);
	
	if (extension->IntDisabled)
	{
		nPort = extension->BaseAddress + 1;
		READ_PORT_UCHAR((PUCHAR)nPort); // clear latch
		
		return TRUE;
	}

	nPort = extension->BaseAddress + 6;
	Value = ((ULONG)READ_PORT_USHORT((PUSHORT)nPort)) >> 4;
	Value |= ((extension->ChannelTable[extension->Index]) << 16);

	extension->DataBuffer[extension->NumPoints - extension->PointsRemaining] = Value;

	extension->PointsRemaining--;
	if (extension->PointsRemaining < 1)
	{
		nPort = extension->BaseAddress + 0xF;
		WRITE_PORT_UCHAR((PUCHAR)nPort, 0xB4); // stop counters
		
		nPort = extension->BaseAddress;
		WRITE_PORT_UCHAR((PUCHAR)nPort, 0);
		extension->IntDisabled = TRUE;
		
		nPort = extension->BaseAddress + 1;
		READ_PORT_UCHAR((PUCHAR)nPort); // clear latch

		return TRUE;
	}

	extension->Index++;
	if (extension->ChannelTable[extension->Index] == -1)
		extension->Index = 0;
	A1216E_SetChannel(extension->ChannelTable[extension->Index], extension);

	nPort = extension->BaseAddress + 1;
	READ_PORT_UCHAR((PUCHAR)nPort); // clear latch 

    return TRUE;
}

/*********************************************************************************
*
*		CANCEL IRP
*
**********************************************************************************/

VOID
A1216ECancelIrp(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp
    )
{
    PA1216E_DEVICE_EXTENSION extension;

    extension = DeviceObject->DeviceExtension;

	Irp->IoStatus.Information = 0;

	Irp->IoStatus.Status = STATUS_CANCELLED;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);
}

